#pragma once

using Genode::memcpy;
using Genode::destroy;

class State_machine::Provider
{
	private:
		State_machine::Env& _env;
		Genode::Xml_node const _config;
		Genode::Signal_context_capability _sigh { };

	protected:
		State_machine::Env& env() { return _env; }
		Genode::Xml_node const& config() { return _config; }

		void notify_sigh()
		{
			if (_sigh.valid()) {
				Genode::Signal_transmitter(_sigh).submit();
			}
		}

	public:
		virtual char const* data() = 0;

		virtual Genode::size_t length() = 0;

		virtual void compute() { }

		void sigh(Genode::Signal_context_capability sigh) 
		{
			_sigh = sigh;
		}

		Provider(State_machine::Env& env, Genode::Xml_node const& config)
		 : _env(env), _config(config)
		{
		}

		virtual ~Provider()
		{
		}
};

class State_machine::Standard_provider : public State_machine::Provider
{
	private:
		struct Data
		{
			char const* data;
			Genode::size_t length;

			Data(char const* data, Genode::size_t length)
			 : data(data), length(length)
			{
			}
		};

		Genode::Constructible<Data> _data { };
		char _dummy = '\0';

	protected:
		void update_data(char const* data, Genode::size_t length)
		{
			if (_data.constructed()) {
				_data.destruct();
			}

			_data.construct(data, length);
		}

	public:
		char const* data() override
		{
			if (_data.constructed()) {
				return _data->data;
			} else {
				return &_dummy;
			}
		}

		Genode::size_t length() override
		{
			if (_data.constructed()) {
				return _data->length;
			} else {
				return 0;
			}
		}

		Standard_provider(State_machine::Env& env, Genode::Xml_node const& config)
		 : Provider(env, config)
		{
		}
};

class State_machine::Buffered_provider : public State_machine::Provider
{
	private:
		class Buffer
		{
			private:
				Genode::Allocator& _alloc;
				Genode::size_t _capacity;
				Genode::size_t _length;
				char* _data;

				Genode::size_t _compute_capacity(Genode::size_t length)
				{
					constexpr Genode::size_t max_length = static_cast<Genode::size_t>(~0ull / 2ull);

					if (length > max_length)
						throw Length_too_large{}; 

					Genode::size_t capacity = 8;

					while (capacity < length) {
						capacity *= 2;
					}

					return capacity;
				}

			public:
				Genode::size_t capacity() { return _capacity; }
				Genode::size_t length() { return _length; }
				char const* data() { return _data; }

				void append(char const* data, Genode::size_t length)
				{
					if (length + _length > _capacity) {
						_capacity = _compute_capacity(length + _length);
						char* new_data = static_cast<char*>(_alloc.alloc(_capacity));
						memcpy(new_data, _data, _length);
						destroy(&_alloc, _data);
						_data = new_data;
					}

					memcpy(_data + _length, data, length);
					_length += length;
				}

				void replace(char const* data, Genode::size_t length)
				{
					if (length > _capacity) {
						destroy(&_alloc, _data);
						_capacity = _compute_capacity(length);
						_data = static_cast<char*>(_alloc.alloc(_capacity));
					}

					memcpy(_data, data, length);
					_length = length;
				}

				Buffer(Genode::Allocator& alloc, Genode::size_t capacity)
				 : _alloc(alloc), _capacity(capacity), _length(0), _data(static_cast<char*>(_alloc.alloc(capacity)))
				{
				}

				~Buffer()
				{
					Genode::destroy(&_alloc, _data);
				}

				Buffer(const Buffer&)            = delete;
				Buffer &operator=(const Buffer&) = delete;
		};

		Buffer _buffer { env().alloc(), 8 };
		char _dummy = '\0';

	protected:
		void set(char const* data, Genode::size_t length)
		{
			_buffer.replace(data, length);
		}

		void append(char const* data, Genode::size_t length)
		{
			_buffer.append(data, length);
		}

	public:
		char const* data() override
		{
			return _buffer.data();
		}

		Genode::size_t length() override
		{
			return _buffer.length();
		}

		Buffered_provider(State_machine::Env& env, Genode::Xml_node const& config)
		 : Provider(env, config)
		{
		}

		Buffered_provider(const Buffered_provider&)            = delete;
		Buffered_provider &operator=(const Buffered_provider&) = delete;
};

class State_machine::Provider_factory
{
	public:
		virtual State_machine::Provider& create(State_machine::Env& env, Genode::Xml_node const& config) = 0;

		virtual ~Provider_factory()
		{
		}
};

